## This file is part of PyANTLR. See LICENSE.txt for license
## details..........Copyright (C) Wolfgang Haefelinger, 2004.

### ANTLR Standard Tokens
SKIP                = -1
INVALID_TYPE        = 0
EOF_TYPE            = 1
NULL_TREE_LOOKAHEAD = 3
MIN_USER_TYPE       = 4


### ANTLR Exceptions
class ANTLRException(Exception):
   pass

class RecognitionException(ANTLRException):
   pass

class NoViableAltException(RecognitionException):
   pass

class NoViableAltForCharException(RecognitionException):
   pass

class SemanticException(RecognitionException):
   pass

class MismatchedCharException(RecognitionException):
   pass

class MismatchedTokenException(RecognitionException):
   pass

class TokenStreamException(ANTLRException):
   pass

class TokenStreamIOException(TokenStreamException):
   pass

class TokenStreamRecognitionException(TokenStreamException):
   pass

class TokenStreamRetryException(TokenStreamException):
   pass

class CharStreamException(ANTLRException):
   pass

class CharStreamIOException(CharStreamException):
   pass


class ArgError(Exception):
   pass

def error(fmt,*args):
   if fmt:
      print "error: ", fmt % args
      
   
class Token(object):

   def __init__(self,**argv):
      try:
         self.type = Token.argv['type']
      except:
         self.type = INVALID_TYPE
      try:
         self.text = argv['text']
      except:
         self.text = "<no text>"

   def isEOF(self):
      return (self.type == EOF_TYPE)
      
   def getColumn(self):
      return 0
   
   def getLine(self):
      return 0

   def getFilename(self):
      return None

   def setFilename(self,name):
      return self
   
   def getText(self):
      return "<no text>"

   def setText(self,text):
      if isinstance(text,str):
         pass
      else:
         raise TypeError
      return self

   def setColumn(self,column):
      return self

   def setLine(self,line):
      return self

   def getType(self):
      return self.type

   def setType(self,type):
      if isinstance(type,int):
         self.type = type
      else:
         raise TypeError
      return self
      
   def toString(self):
      ## not optimal
      type = self.type
      if type == 3:
         tval = 'NULL_TREE_LOOKAHEAD'
      elif type == 1:
         tval = 'EOF_TYPE'
      elif type == 0:
         tval = 'INVALID_TYPE'
      elif type == -1:
         tval = 'SKIP'
      else:
         tval = type
      return '["%s",<%s>]' % (self.getText(),tval)

   def __repr__(self):
      return self.toString()


### static attribute ..
Token.badToken = Token( type=INVALID_TYPE, text="<no text>")


if __name__ == "__main__":
   print "testing .."
   T = Token.badToken
   print T


class CommonToken(Token):

   def __init__(self,**argv):
      Token.__init__(self,**argv)
      self.line = 0
      self.col  = 0
      try:
         self.line = argv['line']
      except:
         pass
      try:
         self.col = argv['col']
      except:
         pass
   
   def getLine(self):
      return self.line

   def getText(self):
      return self.text

   def getColumn(self):
      return self.col

   def setLine(self,line):
      self.line = line
      return self
      
   def setText(self,text):
      self.text = text
      return self

   def setColumn(self,col):
      self.col = col
      return self

   def toString(self):
      ## not optimal 
      d = {
         'text' : self.text,
         'type' : self.type,
         'line' : self.line,
         'colm' : self.col
         }
      fmt = '["%(text)s",<%(type)s>,line=%(line)s,col=%(colm)s]'
      return fmt % d

   def __repr__(self):
      return self.toString()



if __name__ == '__main__' :
   T = CommonToken()
   print T
   T = CommonToken(col=15,line=1,text="some text", type=5)
   print T
   T = CommonToken()
   T.setLine(1).setColumn(15).setText("some text").setType(5)
   print T
   print T.getLine()
   print T.getColumn()
   print T.getText()
   print T.getType()

## Shall be a circular buffer on tokens ..
class CharQueue(object):

   def __init__(self):
      self.buffer = [] # empty list

   def append(self,item):
      self.buffer.append(item)

   def elementAt(self,index):
      return self.buffer[index]

   def reset(self):
      self.buffer = []

   def removeFirst(self):
      self.buffer.pop(0)

   def length(self):
      return len(self.buffer)

   def __str__(self):
      return str(self.buffer)
   

class InputBuffer(object):
   def __init__(self):
      self.nMarkers = 0
      self.markerOffset = 0
      self.numToConsume = 0
      self.queue = CharQueue()

   def __str__(self):
      return "(%s,%s,%s,%s)" % (
         self.nMarkers,
         self.markerOffset,
         self.numToConsume,
         self.queue)

   def __repr__(self):
      return str(self)

   def commit(self):
      self.nMarkers -= 1

   def consume(self) :
      self.numToConsume += 1

   ## probably better to return a list of items
   ## because of unicode. Or return a unicode
   ## string ..
   def getLAChars(self) :
      i = self.markerOffset
      n = self.queue.length()
      s = ''
      while i<n:
         s += self.queue.elementAt(i)
      return s

   ## probably better to return a list of items
   ## because of unicode chars
   def getMarkedChars(self) :
      s = ''
      i = 0
      n = self.markerOffset
      while i<n:
         s += self.queue.elementAt(i)
      return s

   def isMarked(self) :
      return (self.nMarkers != 0)

   def fill(self,k):
      ### abstract method
      assert 0
   
   def LA(self,k) :
      self.fill(k)                        
      return self.queue.elementAt(self.markerOffset + k - 1)

   def mark(self) :
      self.syncConsume()
      self.nMarkers += 1
      return self.markerOffset
   
   def rewind(self,mark) :
      self.syncConsume()
      self.markerOffset = mark
      self.nMarkers -= 1

   def reset(self) :
      self.nMarkers = 0
      self.markerOffset = 0
      self.numToConsume = 0
      self.queue.reset()

   def syncConsume(self) :
      while self.numToConsume > 0:
         if self.nMarkers > 0:
            # guess mode -- leave leading characters and bump offset.
            self.markerOffset += 1
         else:
            # normal mode -- remove first character
            self.queue.removeFirst()
         self.numToConsume -= 1


class CharBuffer(InputBuffer):
   def __init__(self,reader):
      assert isinstance(reader,file)
      super(CharBuffer,self).__init__()
      ## a reader is supposed to be something that has
      ## just a method 'read(int)'.  
      self.input = reader


   def __str__(self):
      base = super(CharBuffer,self).__str__()
      return "CharBuffer{%s,%s}" % (base,str(input))
      
   def fill(self,amount):
      try:
         self.syncConsume()
         while self.queue.length() < (amount + self.markerOffset) :
            ## retrieve just one char - what happend at end
            ## of input?
            char = self.input.read(1)
            self.queue.append(char)
      except Exception,e:
         raise CharStreamIOException(e)
      except:
         error ("unexpected exception caught ..")
         assert 0



class LexerSharedInputState(object):
   def __init__(self,ibuf):
      assert isinstance(ibuf,InputBuffer)
      self.input = ibuf
      self.column = 1
      self.line = 1
      self.tokenStartColumn = 1
      self.tokenStartLine = 1
      self.guessing = 0
      self.filename = None
 
   def reset(self):
      self.column = 1
      self.line = 1
      self.tokenStartColumn = 1
      self.tokenStartLine = 1
      self.guessing = 0
      self.filename = None
      self.input.reset()


   def LA(self,k):
     return self.input.LA(k)


class TokenStream(object):
   def nextToken(self):
      pass


class StringBuffer:
   def __init__(self,string=None):
      if string:
         self.text = list(string)
      else:
         self.text = []

   def setLength(self,len):
      assert len == 0
      self.text = []

   def length(self):
      return len(self.text)

   def append(self,c):
      self.text.append(c)

   ### return buffer as string. Arg 'a' is  used  as index
   ## into the buffer and 2nd argument shall be the length.
   ## If 2nd args is absent, we return chars till end of
   ## buffer starting with 'a'.
   def getString(self,a=None,length=None):
      if not a:
         a = 0
      assert a>=0 and a<len(self.text)
      if not length:
         if length == None:
            ## no second argument
            L = self.text[a:]
         else:
            return ""
      else:
         assert (a+length) <= len(self.text)
         b = a + length
         L = self.text[a:b]
      s = ""
      for x in L : s += x
      return s
   


class CharScanner(TokenStream):
   ## class members
   NO_CHAR = 0
   EOF_CHAR = -1

   def __init__(self,*argv) :
      super(CharScanner,self).__init__()
      self.saveConsumedInput = True
      self.tokenObjectClass = None
      self.caseSensitive = True
      self.caseSensitiveLiterals = True
      self.literals = None
      self.tabsize = 8
      self._returnToken = None
      self.commitToPath = False
      self.traceDepth = 0
      self.text = StringBuffer()
      self.hashString = hash(self)
      self.setTokenObjectClass("antlr.CommonToken")
      self.setInput(*argv)

   def setInput(self,*argv):
      ## case 1:
      ## if there's no arg we default to read from
      ## standard input
      if not argv:
         import sys
         self.setInput(sys.stdin)
         return

      ## get 1st argument
      arg1 = argv[0]
      
      ## case 2:
      ## if arg1 is a string,  we assume it's a file name
      ## and  open  a  stream  using 2nd argument as open
      ## mode. If there's no 2nd argument we fall back to
      ## mode '+rb'.
      if isinstance(arg1,str):
         f = open(arg1,"rb")
         self.setInput(f)
         self.setFilename(arg1)
         return

      ## case 3:
      ## if arg1 is a file we wrap it by a char buffer (
      ## some additional checks?? No, can't do this in
      ## general).
      if isinstance(arg1,file):
         self.setInput(CharBuffer(arg1))
         return

      ## case 4:
      ## if arg1 is of type SharedLexerInputState we use
      ## argument as is.
      if isinstance(arg1,LexerSharedInputState):
         self.inputState = arg1
         return

      ## case 5:
      ## check whether argument type is of type input
      ## buffer. If so create a SharedLexerInputState and
      ## go ahead.
      if isinstance(arg1,InputBuffer):
         self.setInput(LexerSharedInputState(arg1))
         return

      ## case 6:
      ## check whether argument type has a method read(int)
      ## If so create CharBuffer ...
      ## (not yet implemented)

      ## case 7:
      ## raise wrong argument exception
      raise ArgError(argv)

   def setTabSize(self,size) :
      self.tabsize = size

   def getTabSize(self) :
      return self.tabsize

   def setCaseSensitive(self,t) :
       self.caseSensitive = t


   def setCommitToPath(self,commit) : 
      self.commitToPath = commit


   def setFilename(self,f) : 
      self.inputState.filename = f


   def setLine(self,line) :
      self.inputState.line = line


   def setText(self,s) :
      self.resetText()
      self.text.append(s)


   def getCaseSensitive(self) :
      return self.caseSensitive


   def getCaseSensitiveLiterals(self) :
      return self.caseSensitiveLiterals


   def getColumn(self) :
      return self.inputState.column


   def setColumn(self,c) :
      self.inputState.column = c


   def getCommitToPath(self) :
      return self.commitToPath


   def getFilename(self) :
      return self.inputState.filename


   def getInputBuffer(self) :
      return self.inputState.input


   def getInputState(self) :
      return self.inputState


   def setInputState(self,state) :
      assert isinstance(state,LexerSharedInputState)
      self.inputState = state


   def getLine(self) :
      return self.inputState.line


   def getText(self) :
      return self.text.toString()


   def getTokenObject(self) :
      return self._returnToken


   def LA(self,i) :
      c = self.inputState.input.LA(i)
      if not self.caseSensitive:
         c = str.lower(c)
      return c

   def makeToken(self,type) :
      try:
         ## how to dynamically load a class?
         tok = CommonToken()
         tok.setType(type)
         tok.setColumn(self.inputState.tokenStartColumn)
         tok.setLine(self.inputState.tokenStartLine)
         return tok
      except:
         panic("unable to create new token")
      return Token.badToken

   def mark(self) :
      return self.inputState.input.mark()


   def _match_bitset(self,b) :
      if b.member(self.LA(1)):
         self.consume()
      else:
         raise MismatchedCharException(self.LA(1), b, False,self)

   def _match_string(self,s) :
      for c in s:
         if self.LA(1) == c:
            self.consume()
         else:
            raise MismatchedCharException(self.LA(1),c,False,self)

   def match(self,item):
      if isinstance(item,str) or isinstance(item,unicode):
         return self._match_string(item)
      else:
         return self._match_bitset(item)

         
   def matchNot(self,c) :
      if self.LA(1) != c:
         self.consume()
      else:
         raise MismatchedCharException(self.LA(1), c, True,self)

   def matchRange(self,c1,c2) :
      if self.LA(1) < c1 or self.LA(1) > c2 :
         raise MismatchedCharException(self.LA(1), c1, c2, False,self)
      else:
         self.consume()

   def newline(self) :
      self.inputState.line += 1
      self.inputState.column = 1
      

   def tab(self) :
      c = self.getColumn()
      nc = ( ((c-1)/self.tabsize) + 1) * self.tabsize + 1
      self.setColumn(nc)
      

   def panic(self,s='') :
      print "CharScanner: panic: " + s 
      import sys
      sys.exit(1)
      
      
   def reportError(self,ex) :
      print ex
      
      
   def reportError(self,s) :
      if not self.getFilename():
         print "error: " + s
      else:
         print self.getFilename() + ": error: " + s
         
   def reportWarning(self,s) :
      if not self.getFilename():
         print "warning: " + s
      else:
         print self.getFilename() + ": warning: " + s


   def resetText(self) :
      self.text.setLength(0)
      self.inputState.tokenStartColumn = self.inputState.column
      self.inputState.tokenStartLine = self.inputState.line
      

   def rewind(self,pos) :
      self.inputState.input.rewind(pos)
      

   def setTokenObjectClass(self,cl):
      self.tokenClass = cl

   def testForLiteral(self,token):
      assert token
      assert isinstance(token,Token)

      _type = token.getType()

      ## special tokens can't be literals
      if _type in [SKIP,INVALID_TYPE,EOF_TYPE,NULL_TREE_LOOKAHEAD] :
         return

      _text = token.getText()
      assert _text

      _type = self.testLiteralsTable(_text,_type)
      token.setType(_type)


   def testLiteralsTable(self,*args):
      if isinstance(args[0],str):
         s = args[0]
         i = args[1]
      else:
         s = self.text.getString()
         i = args[0]

      ## check whether integer has been given
      assert isinstance(i,int)

      ## check whether we have a dict
      assert isinstance(self.literals,dict)
      try:
         i = self.literals[s]
      except:
         pass
      return i
   
   def toLower(self,c):
      return c.lower()

   def traceIndent(self):
      print ' ' * self.traceDepth

   def traceIn(self,rname):
      self.traceDepth += 1
      self.traceIndent()
      print "> lexer %s c== %s" % (rname,self.LA(1))


   def traceOut(self,rname):
      self.traceIndent()
      print "< lexer %s c== %s" % (rname,self.LA(1))
      self.traceDepth -= 1

   def uponEOF(self):
      pass


   def append(self,c):
      if self.saveConsumedInput :
         self.text.append(c)

   def commit(self):
      self.inputState.input.commit()

   def consume(self):
      if self.inputState.guessing == 0:
         c = self.LA(1)
         if self.caseSensitive:
            self.append(c)
         else:
            # use input.LA(), not LA(), to get original case
            # CharScanner.LA() would toLower it.
            append(self.inputState.input.LA(1))

         if c in "\t":
            self.tab()
         else:
            self.inputState.column += 1
      self.inputState.input.consume()

   ## Consume chars until one matches the given char
   def consumeUntil_char(self,c):
      while self.LA(1) != EOF_CHAR and self.LA(1) != c: 
         self.consume()

   ## Consume chars until one matches the given set
   def consumeUntil_bitset(self,bitset):
      while self.LA(1) != EOF_CHAR and not self.set.member(self.LA(1)):
         self.consume()

   ### If symbol seen is EOF then generate and set token, otherwise
   ### throw exception.
   def lasthopeEOF(self,la1):
      if not la1:
         self.uponEOF()
         self._returnToken = self.makeToken(EOF_TYPE)
      else:
         self.raise_NoViableAlt(la1)

   def raise_NoViableAlt(self,la1=None):
      if not la1: la1 = self.LA(1)
      fname = self.getFilename()
      line  = self.getLine()
      col   = self.getColumn()
      raise NoViableAltForCharException(la1,fname,line,col)

   def set_return_token(self,_create,_token,_ttype,_offset):
      if _create and _token==None and (not _ttype == SKIP):
         string = self.text.getString(_offset)
         _token = self.makeToken(_ttype)
         _token.setText(string)
      self._returnToken = _token
      return _token



### BitSet class
class BitSet(object):
   BITS     = 64
   NIBBLE   = 4
   LOG_BITS = 6
   MOD_MASK = BITS -1

   
   def __init__(self,data):
      if not data:
         raise ArgError(self)
      if not isinstance(data,list):
         raise ArgError(self,data)

      for x in data:
         if not isinstance(x,long):
            raise ArgError(self,"this is not a long: %s" % (x))
      self.data = data


   def __str__(self):
      return str(self.data)

   def __repr__(self):
      return str(self)


   def member(self,char):
      if not char:
         return False
      
      ## char is a (unicode) string with at most lenght 1, ie.
      ## a char.
      
      if len(char) != 1:
         raise ArgError(self,"char expected: %s" % (char))

      if not (isinstance(char,str) or isinstance(char,unicode)):
          raise ArgError(self,"char or unichar expected: %s" % (char))

      ### handle ASCII/UNICODE char
      num = ord(char)

      ### check whether position num is in bitset
      return self.at(num)


   def wordNumber(self,bit):
      return bit >> BitSet.LOG_BITS

   def bitMask(self,bit):
      pos = bit & BitSet.MOD_MASK  ## bit mod BITS
      return (1L << pos)

   def set(self,bit,on=True):
      i = self.wordNumber(bit)
      mask = self.bitMask(bit)
      if on:
         self.data[i] |=  mask
      else:
         self.data[i] &= (~mask)

   def off(self,bit,off=True):
      self.set(bit,not off)
      

   def at(self,bit):
      i = self.wordNumber(bit)
      v = self.data[i]
      m = self.bitMask(bit)
      return v & m
   
      


      
      
      
